/* Copyright (C) 2009-2011 Kentoku Shiba

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#define MYSQL_SERVER 1
#include "mysql_version.h"
#if MYSQL_VERSION_ID < 50500
#include "mysql_priv.h"
#include <mysql/plugin.h>
#else
#include "sql_priv.h"
#include "probes_mysql.h"
#include "sql_class.h"
#include "sql_base.h"
#endif
#ifdef WITH_PARTITION_STORAGE_ENGINE
#include "ha_partition.h"
#endif
#include "vp_err.h"
#include "vp_param.h"
#include "vp_include.h"
#include "ha_vp.h"
#include "vp_table.h"

handlerton *vp_hton_ptr;

#ifdef HAVE_PSI_INTERFACE
PSI_mutex_key vp_key_mutex_tbl;
#ifdef WITH_PARTITION_STORAGE_ENGINE
PSI_mutex_key vp_key_mutex_pt_share;
#endif
PSI_mutex_key vp_key_mutex_bg_sync;
PSI_mutex_key vp_key_mutex_bg;
PSI_mutex_key vp_key_mutex_share;
PSI_mutex_key vp_key_mutex_share_init;
#ifdef WITH_PARTITION_STORAGE_ENGINE
PSI_mutex_key vp_key_mutex_pt_handler;
#endif

static PSI_mutex_info all_vp_mutexes[]=
{
  { &vp_key_mutex_tbl, "tbl", PSI_FLAG_GLOBAL},
#ifdef WITH_PARTITION_STORAGE_ENGINE
  { &vp_key_mutex_pt_share, "pt_share", PSI_FLAG_GLOBAL},
#endif
#ifndef WITHOUT_VP_BG_ACCESS
  { &vp_key_mutex_bg_sync, "bg_sync", 0},
  { &vp_key_mutex_bg, "bg", 0},
#endif
  { &vp_key_mutex_share, "share", 0},
  { &vp_key_mutex_share_init, "share_init", 0},
#ifdef WITH_PARTITION_STORAGE_ENGINE
  { &vp_key_mutex_pt_handler, "pt_handler", 0},
#endif
};

#ifndef WITHOUT_VP_BG_ACCESS
PSI_cond_key vp_key_cond_bg_sync;
PSI_cond_key vp_key_cond_bg;
#endif

static PSI_cond_info all_vp_conds[] = {
#ifndef WITHOUT_VP_BG_ACCESS
  {&vp_key_cond_bg_sync, "bg_sync", 0},
  {&vp_key_cond_bg, "bg", 0},
#endif
};

#ifndef WITHOUT_VP_BG_ACCESS
PSI_thread_key vp_key_thd_bg;
#endif

static PSI_thread_info all_vp_threads[] = {
#ifndef WITHOUT_VP_BG_ACCESS
  {&vp_key_thd_bg, "bg", 0},
#endif
};
#endif

HASH vp_open_tables;
pthread_mutex_t vp_tbl_mutex;

#ifdef WITH_PARTITION_STORAGE_ENGINE
HASH vp_open_pt_share;
pthread_mutex_t vp_pt_share_mutex;
#endif

#ifndef WITHOUT_VP_BG_ACCESS
pthread_attr_t vp_pt_attr;
#endif

// for vp_open_tables
uchar *vp_tbl_get_key(
  VP_SHARE *share,
  size_t *length,
  my_bool not_used __attribute__ ((unused))
) {
  DBUG_ENTER("vp_tbl_get_key");
  *length = share->table_name_length;
  DBUG_RETURN((uchar*) share->table_name);
}

#ifdef WITH_PARTITION_STORAGE_ENGINE
uchar *vp_pt_share_get_key(
  VP_PARTITION_SHARE *share,
  size_t *length,
  my_bool not_used __attribute__ ((unused))
) {
  DBUG_ENTER("vp_pt_share_get_key");
  *length = share->table_name_length;
  DBUG_RETURN((uchar*) share->table_name);
}

uchar *vp_pt_handler_share_get_key(
  VP_PARTITION_HANDLER_SHARE *share,
  size_t *length,
  my_bool not_used __attribute__ ((unused))
) {
  DBUG_ENTER("vp_pt_handler_share_get_key");
  *length = sizeof(TABLE *);
  DBUG_RETURN((uchar*) &share->table);
}
#endif

#ifdef HAVE_PSI_INTERFACE
static void init_vp_psi_keys()
{
  DBUG_ENTER("init_vp_psi_keys");
  if (PSI_server == NULL)
    DBUG_VOID_RETURN;

  PSI_server->register_mutex("vp", all_vp_mutexes,
    array_elements(all_vp_mutexes));
  PSI_server->register_cond("vp", all_vp_conds,
    array_elements(all_vp_conds));
  PSI_server->register_thread("vp", all_vp_threads,
    array_elements(all_vp_threads));
  DBUG_VOID_RETURN;
}
#endif

int vp_free_share_alloc(
  VP_SHARE *share
) {
  DBUG_ENTER("vp_free_share_alloc");
  if (share->tgt_default_db_name)
    my_free(share->tgt_default_db_name, MYF(0));
  if (share->tgt_table_name_list)
    my_free(share->tgt_table_name_list, MYF(0));
  if (share->tgt_table_name_prefix)
    my_free(share->tgt_table_name_prefix, MYF(0));
  if (share->tgt_table_name_suffix)
    my_free(share->tgt_table_name_suffix, MYF(0));
  if (share->choose_ignore_table_list)
    my_free(share->choose_ignore_table_list, MYF(0));
  if (share->choose_ignore_table_list_for_lock)
    my_free(share->choose_ignore_table_list_for_lock, MYF(0));
  if (share->tgt_db_name)
    my_free(share->tgt_db_name, MYF(0));
  if (share->correspond_columns_p)
    my_free(share->correspond_columns_p, MYF(0));
#ifdef WITH_PARTITION_STORAGE_ENGINE
  if (share->partition_share)
    vp_free_pt_share(share->partition_share);
#endif
  DBUG_RETURN(0);
}

char *vp_get_string_between_quote(
  char *ptr,
  bool alloc
) {
  char *start_ptr, *end_ptr, *tmp_ptr, *esc_ptr;
  bool find_flg = FALSE, esc_flg = FALSE;
  DBUG_ENTER("vp_get_string_between_quote");

  start_ptr = strchr(ptr, '\'');
  end_ptr = strchr(ptr, '"');
  if (start_ptr && (!end_ptr || start_ptr < end_ptr))
  {
    tmp_ptr = ++start_ptr;
    while (!find_flg)
    {
      if (!(end_ptr = strchr(tmp_ptr, '\'')))
        DBUG_RETURN(NULL);
      esc_ptr = tmp_ptr;
      while (!find_flg)
      {
        esc_ptr = strchr(esc_ptr, '\\');
        if (!esc_ptr || esc_ptr > end_ptr)
          find_flg = TRUE;
        else if (esc_ptr == end_ptr - 1)
        {
          esc_flg = TRUE;
          tmp_ptr = end_ptr + 1;
          break;
        } else {
          esc_flg = TRUE;
          esc_ptr += 2;
        }
      }
    }
  } else if (end_ptr)
  {
    start_ptr = end_ptr;
    tmp_ptr = ++start_ptr;
    while (!find_flg)
    {
      if (!(end_ptr = strchr(tmp_ptr, '"')))
        DBUG_RETURN(NULL);
      esc_ptr = tmp_ptr;
      while (!find_flg)
      {
        esc_ptr = strchr(esc_ptr, '\\');
        if (!esc_ptr || esc_ptr > end_ptr)
          find_flg = TRUE;
        else if (esc_ptr == end_ptr - 1)
        {
          esc_flg = TRUE;
          tmp_ptr = end_ptr + 1;
          break;
        } else {
          esc_flg = TRUE;
          esc_ptr += 2;
        }
      }
    }
  } else
    DBUG_RETURN(NULL);

  *end_ptr = '\0';
  if (esc_flg)
  {
    esc_ptr = start_ptr;
    while (TRUE)
    {
      esc_ptr = strchr(esc_ptr, '\\');
      if (!esc_ptr)
        break;
      switch(*(esc_ptr + 1))
      {
        case 'b':
          *esc_ptr = '\b';
          break;
        case 'n':
          *esc_ptr = '\n';
          break;
        case 'r':
          *esc_ptr = '\r';
          break;
        case 't':
          *esc_ptr = '\t';
          break;
        default:
          *esc_ptr = *(esc_ptr + 1);
          break;
      }
      esc_ptr++;
      strcpy(esc_ptr, esc_ptr + 1);
    }
  }
  if (alloc)
  {
    DBUG_RETURN(
      vp_create_string(
      start_ptr,
      strlen(start_ptr))
    );
  } else {
    DBUG_RETURN(start_ptr);
  }
}

#define VP_PARAM_STR_LEN(name) name ## _length
#define VP_PARAM_STR(title_name, param_name) \
  if (!strncasecmp(tmp_ptr, title_name, title_length)) \
  { \
    DBUG_PRINT("info",("vp "title_name" start")); \
    if (!share->param_name) \
    { \
      if ((share->param_name = vp_get_string_between_quote( \
        start_ptr, TRUE))) \
        share->VP_PARAM_STR_LEN(param_name) = strlen(share->param_name); \
      else { \
        error_num = ER_VP_INVALID_TABLE_INFO_NUM; \
        my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR, \
          MYF(0), tmp_ptr); \
        goto error; \
      } \
      DBUG_PRINT("info",("vp "title_name"=%s", share->param_name)); \
    } \
    break; \
  }
#define VP_PARAM_INT_WITH_MAX(title_name, param_name, min_val, max_val) \
  if (!strncasecmp(tmp_ptr, title_name, title_length)) \
  { \
    DBUG_PRINT("info",("vp "title_name" start")); \
    if (share->param_name == -1) \
    { \
      if ((tmp_ptr2 = vp_get_string_between_quote( \
        start_ptr, FALSE))) \
      { \
        share->param_name = atoi(tmp_ptr2); \
        if (share->param_name < min_val) \
          share->param_name = min_val; \
        else if (share->param_name > max_val) \
          share->param_name = max_val; \
      } else { \
        error_num = ER_VP_INVALID_TABLE_INFO_NUM; \
        my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR, \
          MYF(0), tmp_ptr); \
        goto error; \
      } \
      DBUG_PRINT("info",("vp "title_name"=%d", share->param_name)); \
    } \
    break; \
  }

int vp_parse_table_info(
  VP_SHARE *share,
  TABLE *table,
  uint create_table
) {
  int error_num = 0;
  char *comment_string = NULL;
  char *sprit_ptr[2];
  char *tmp_ptr, *tmp_ptr2, *start_ptr;
  int roop_count;
  int title_length;
#ifdef WITH_PARTITION_STORAGE_ENGINE
  partition_element *part_elem;
  partition_element *sub_elem;
#endif
  DBUG_ENTER("vp_parse_table_info");
#if MYSQL_VERSION_ID < 50500
  DBUG_PRINT("info",("vp partition_info=%s", table->s->partition_info));
#else
  DBUG_PRINT("info",("vp partition_info=%s", table->s->partition_info_str));
#endif
  DBUG_PRINT("info",("vp part_info=%x", table->part_info));
  DBUG_PRINT("info",("vp s->db=%s", table->s->db.str));
  DBUG_PRINT("info",("vp s->table_name=%s", table->s->table_name.str));
  DBUG_PRINT("info",("vp s->path=%s", table->s->path.str));
  DBUG_PRINT("info",
    ("vp s->normalized_path=%s", table->s->normalized_path.str));
#ifdef WITH_PARTITION_STORAGE_ENGINE
  vp_get_partition_info(share->table_name, table, &part_elem, &sub_elem);
#endif
  share->choose_table_mode = -1;
  share->choose_table_mode_for_lock = -1;
  share->multi_range_mode = -1;
  share->pk_correspond_mode = -1;
  share->info_src_table = -1;
  share->auto_increment_table = -1;
  share->table_count_mode = -1;
  share->support_table_cache = -1;
  share->child_binlog = -1;
#ifndef WITHOUT_VP_BG_ACCESS
  share->bgs_mode = -1;
  share->bgi_mode = -1;
  share->bgu_mode = -1;
#endif
  share->zero_record_update_mode = -1;

#ifdef WITH_PARTITION_STORAGE_ENGINE
  for (roop_count = 4; roop_count > 0; roop_count--)
#else
  for (roop_count = 2; roop_count > 0; roop_count--)
#endif
  {
    if (comment_string)
    {
      my_free(comment_string, MYF(0));
      comment_string = NULL;
    }
    switch (roop_count)
    {
#ifdef WITH_PARTITION_STORAGE_ENGINE
      case 4:
        if (!sub_elem || !sub_elem->part_comment)
          continue;
        DBUG_PRINT("info",("vp create sub comment string"));
        if (
          !(comment_string = vp_create_string(
            sub_elem->part_comment,
            strlen(sub_elem->part_comment)))
        ) {
          error_num = HA_ERR_OUT_OF_MEM;
          goto error_alloc_comment_string;
        }
        DBUG_PRINT("info",("vp sub comment string=%s", comment_string));
        break;
      case 3:
        if (!part_elem || !part_elem->part_comment)
          continue;
        DBUG_PRINT("info",("vp create part comment string"));
        if (
          !(comment_string = vp_create_string(
            part_elem->part_comment,
            strlen(part_elem->part_comment)))
        ) {
          error_num = HA_ERR_OUT_OF_MEM;
          goto error_alloc_comment_string;
        }
        DBUG_PRINT("info",("vp part comment string=%s", comment_string));
        break;
#endif
      case 2:
        if (table->s->comment.length == 0)
          continue;
        DBUG_PRINT("info",("vp create comment string"));
        if (
          !(comment_string = vp_create_string(
            table->s->comment.str,
            table->s->comment.length))
        ) {
          error_num = HA_ERR_OUT_OF_MEM;
          goto error_alloc_comment_string;
        }
        DBUG_PRINT("info",("vp comment string=%s", comment_string));
        break;
      default:
        if (table->s->connect_string.length == 0)
          continue;
        DBUG_PRINT("info",("vp create connect_string string"));
        if (
          !(comment_string = vp_create_string(
            table->s->connect_string.str,
            table->s->connect_string.length))
        ) {
          error_num = HA_ERR_OUT_OF_MEM;
          goto error_alloc_comment_string;
        }
        DBUG_PRINT("info",("vp comment_string=%s", comment_string));
        break;
    }

    sprit_ptr[0] = comment_string;
    while (sprit_ptr[0])
    {
      if ((sprit_ptr[1] = strchr(sprit_ptr[0], ',')))
      {
        *sprit_ptr[1] = '\0';
        sprit_ptr[1]++;
      }
      tmp_ptr = sprit_ptr[0];
      sprit_ptr[0] = sprit_ptr[1];
      while (*tmp_ptr == ' ' || *tmp_ptr == '\r' ||
        *tmp_ptr == '\n' || *tmp_ptr == '\t')
        tmp_ptr++;

      if (*tmp_ptr == '\0')
        continue;

      title_length = 0;
      start_ptr = tmp_ptr;
      while (*start_ptr != ' ' && *start_ptr != '\'' &&
        *start_ptr != '"' && *start_ptr != '\0' &&
        *start_ptr != '\r' && *start_ptr != '\n' &&
        *start_ptr != '\t')
      {
        title_length++;
        start_ptr++;
      }

      switch (title_length)
      {
        case 0:
          continue;
        case 3:
          VP_PARAM_INT_WITH_MAX("ait", auto_increment_table, 1,
            share->table_count);
#ifndef WITHOUT_VP_BG_ACCESS
          VP_PARAM_INT_WITH_MAX("bgs", bgs_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("bgi", bgi_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("bgu", bgu_mode, 0, 1);
#endif
          VP_PARAM_INT_WITH_MAX("cbl", child_binlog, 0, 1);
          VP_PARAM_STR("cil", choose_ignore_table_list_for_lock);
          VP_PARAM_STR("cit", choose_ignore_table_list);
          VP_PARAM_INT_WITH_MAX("cml", choose_table_mode_for_lock, 0, 1);
          VP_PARAM_INT_WITH_MAX("ctm", choose_table_mode, 0, 1);
          VP_PARAM_STR("ddb", tgt_default_db_name);
          VP_PARAM_INT_WITH_MAX("tcm", table_count_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("ist", info_src_table, 0, share->table_count);
          VP_PARAM_INT_WITH_MAX("mrm", multi_range_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("pcm", pk_correspond_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("stc", support_table_cache, 0, 2);
          VP_PARAM_STR("tnl", tgt_table_name_list);
          VP_PARAM_STR("tnp", tgt_table_name_prefix);
          VP_PARAM_STR("tns", tgt_table_name_suffix);
          VP_PARAM_INT_WITH_MAX("zru", zero_record_update_mode, 0, 1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
#ifndef WITHOUT_VP_BG_ACCESS
        case 8:
          VP_PARAM_INT_WITH_MAX("bgs_mode", bgs_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("bgi_mode", bgi_mode, 0, 1);
          VP_PARAM_INT_WITH_MAX("bgu_mode", bgu_mode, 0, 1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
#endif
        case 12:
          VP_PARAM_INT_WITH_MAX("child_binlog", child_binlog, 0, 1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 15:
          VP_PARAM_STR("table_name_list", tgt_table_name_list);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 16:
          VP_PARAM_INT_WITH_MAX("multi_range_mode", multi_range_mode, 0, 1);
          VP_PARAM_STR("default_database", tgt_default_db_name);
          VP_PARAM_INT_WITH_MAX("table_count_mode", table_count_mode, 0, 1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 17:
          VP_PARAM_INT_WITH_MAX("choose_table_mode", choose_table_mode, 0, 1);
          VP_PARAM_STR("table_name_prefix", tgt_table_name_prefix);
          VP_PARAM_STR("table_name_suffix", tgt_table_name_suffix);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 18:
          VP_PARAM_INT_WITH_MAX("pk_correspond_mode", pk_correspond_mode, 0,
            1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 19:
          VP_PARAM_INT_WITH_MAX("support_table_cache", support_table_cache, 0,
            2);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 20:
          VP_PARAM_INT_WITH_MAX("auto_increment_table", auto_increment_table,
            1, share->table_count);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 23:
          VP_PARAM_INT_WITH_MAX("infomation_source_table", info_src_table, 0,
            share->table_count);
          VP_PARAM_INT_WITH_MAX("zero_record_update_mode",
            zero_record_update_mode, 0, 1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 24:
          VP_PARAM_STR("choose_ignore_table_list", choose_ignore_table_list);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 26:
          VP_PARAM_INT_WITH_MAX("choose_table_mode_for_lock",
            choose_table_mode_for_lock, 0, 1);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        case 33:
          VP_PARAM_STR("choose_ignore_table_list_for_lock",
            choose_ignore_table_list_for_lock);
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
        default:
          error_num = ER_VP_INVALID_TABLE_INFO_NUM;
          my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_STR,
            MYF(0), tmp_ptr);
          goto error;
      }
    }
  }

  if ((error_num = vp_set_table_info_default(
    share,
#ifdef WITH_PARTITION_STORAGE_ENGINE
    part_elem,
    sub_elem,
#endif
    table
  )))
    goto error;

  if (create_table)
  {
    DBUG_PRINT("info",
      ("vp tgt_default_db_name_length = %ld",
      share->tgt_default_db_name_length));
    if (share->tgt_default_db_name_length > VP_TABLE_INFO_MAX_LEN)
    {
      error_num = ER_VP_INVALID_TABLE_INFO_TOO_LONG_NUM;
      my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_TOO_LONG_STR,
        MYF(0), share->tgt_default_db_name, "default_database");
      goto error;
    }

/*
    DBUG_PRINT("info",
      ("vp tgt_table_name_list_length = %ld",
      share->tgt_table_name_list_length));
    if (share->tgt_table_name_list_length > VP_TABLE_INFO_MAX_LEN)
    {
      error_num = ER_VP_INVALID_TABLE_INFO_TOO_LONG_NUM;
      my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_TOO_LONG_STR,
        MYF(0), share->tgt_table_name_list, "table_name_list");
      goto error;
    }
*/

    DBUG_PRINT("info",
      ("vp tgt_table_name_prefix_length = %ld",
      share->tgt_table_name_prefix_length));
    if (share->tgt_table_name_prefix_length > VP_TABLE_INFO_MAX_LEN)
    {
      error_num = ER_VP_INVALID_TABLE_INFO_TOO_LONG_NUM;
      my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_TOO_LONG_STR,
        MYF(0), share->tgt_table_name_prefix, "table_name_prefix");
      goto error;
    }

    DBUG_PRINT("info",
      ("vp tgt_table_name_suffix_length = %ld",
      share->tgt_table_name_suffix_length));
    if (share->tgt_table_name_suffix_length > VP_TABLE_INFO_MAX_LEN)
    {
      error_num = ER_VP_INVALID_TABLE_INFO_TOO_LONG_NUM;
      my_printf_error(error_num, ER_VP_INVALID_TABLE_INFO_TOO_LONG_STR,
        MYF(0), share->tgt_table_name_suffix, "table_name_suffix");
      goto error;
    }
  }

  if (comment_string)
    my_free(comment_string, MYF(0));
  DBUG_RETURN(0);

error:
  if (comment_string)
    my_free(comment_string, MYF(0));
error_alloc_comment_string:
  DBUG_RETURN(error_num);
}

int vp_set_table_info_default(
  VP_SHARE *share,
#ifdef WITH_PARTITION_STORAGE_ENGINE
  partition_element *part_elem,
  partition_element *sub_elem,
#endif
  TABLE *table
) {
  DBUG_ENTER("vp_set_table_info_default");
  if (!share->tgt_default_db_name)
  {
    DBUG_PRINT("info",("vp create default tgt_default_db_name"));
    share->tgt_default_db_name_length = table->s->db.length;
    if (
      !(share->tgt_default_db_name = vp_create_string(
        table->s->db.str,
        share->tgt_default_db_name_length))
    ) {
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    }
  }

  if (!share->tgt_table_name_prefix)
  {
    DBUG_PRINT("info",("vp create default tgt_table_name_prefix"));
    share->tgt_table_name_prefix_length = 0;
    if (
      !(share->tgt_table_name_prefix = vp_create_string(
        "",
        share->tgt_table_name_prefix_length))
    ) {
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    }
  }

  if (!share->tgt_table_name_suffix)
  {
    DBUG_PRINT("info",("vp create default tgt_table_name_suffix"));
    share->tgt_table_name_suffix_length = 0;
    if (
      !(share->tgt_table_name_suffix = vp_create_string(
        "",
        share->tgt_table_name_suffix_length))
    ) {
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    }
  }

  if (!share->tgt_table_name_list && table)
  {
    DBUG_PRINT("info",("vp create default tgt_table_name_list"));
    share->tgt_table_name_list_length = share->table_name_length;
    if (
      !(share->tgt_table_name_list = vp_create_table_name_string(
        table->s->table_name.str,
#ifdef WITH_PARTITION_STORAGE_ENGINE
        (part_elem ? part_elem->partition_name : NULL),
        (sub_elem ? sub_elem->partition_name : NULL)
#else
        NULL,
        NULL
#endif
      ))
    ) {
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    }
  }

  if (!share->choose_ignore_table_list)
  {
    DBUG_PRINT("info",("vp create default choose_ignore_table_list"));
    share->choose_ignore_table_list_length = 0;
    if (
      !(share->choose_ignore_table_list = vp_create_string(
        "",
        share->choose_ignore_table_list_length))
    ) {
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    }
  }

  if (!share->choose_ignore_table_list_for_lock)
  {
    DBUG_PRINT("info",("vp create default choose_ignore_table_list_for_lock"));
    share->choose_ignore_table_list_for_lock_length = 0;
    if (
      !(share->choose_ignore_table_list_for_lock = vp_create_string(
        "",
        share->choose_ignore_table_list_for_lock_length))
    ) {
      DBUG_RETURN(HA_ERR_OUT_OF_MEM);
    }
  }

  if (share->choose_table_mode == -1)
    share->choose_table_mode = 0;
  if (share->choose_table_mode_for_lock == -1)
    share->choose_table_mode_for_lock = 1;
  if (share->multi_range_mode == -1)
    share->multi_range_mode = 1;
  if (share->pk_correspond_mode == -1)
    share->pk_correspond_mode = 0;
  if (share->info_src_table == -1)
    share->info_src_table = 0;
  if (share->table_count_mode == -1)
    share->table_count_mode = 0;
  if (share->support_table_cache == -1)
    share->support_table_cache = 1;
  if (share->auto_increment_table == -1)
    share->auto_increment_table = 1;
  if (share->child_binlog == -1)
    share->child_binlog = 0;
#ifndef WITHOUT_VP_BG_ACCESS
  if (share->bgs_mode == -1)
    share->bgs_mode = 0;
  if (share->bgi_mode == -1)
    share->bgi_mode = 0;
  if (share->bgu_mode == -1)
    share->bgu_mode = 0;
#endif
  if (share->zero_record_update_mode == -1)
    share->zero_record_update_mode = 0;
  DBUG_RETURN(0);
}

VP_SHARE *vp_get_share(
  const char *table_name,
  TABLE *table,
  const THD *thd,
  ha_vp *vp,
  int *error_num
) {
  VP_SHARE *share;
  uint length;
  char *tmp_name;
  DBUG_ENTER("vp_get_share");

  length = (uint) strlen(table_name);
  pthread_mutex_lock(&vp_tbl_mutex);
  if (!(share = (VP_SHARE*) my_hash_search(&vp_open_tables,
    (uchar*) table_name, length)))
  {
    DBUG_PRINT("info",("vp create new share"));
    if (!(share = (VP_SHARE *)
      my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
        &share, sizeof(*share),
        &tmp_name, length + 1,
        NullS))
    ) {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_alloc_share;
    }

    share->use_count = 0;
    share->table_name_length = length;
    share->table_name = tmp_name;
    strmov(share->table_name, table_name);

    if ((*error_num = vp_parse_table_info(share, table, 0)))
      goto error_parse_comment_string;

    if (
      (*error_num = vp_create_table_list(share)) ||
      (*error_num = vp_table_num_list_to_bitmap(share,
        share->choose_ignore_table_list, share->select_ignore)) ||
      (*error_num = vp_table_num_list_to_bitmap(share,
        share->choose_ignore_table_list_for_lock,
        share->select_ignore_with_lock))
    )
      goto error_create_table_list;

    if (share->table_count_mode)
      share->additional_table_flags |= HA_STATS_RECORDS_IS_EXACT;

    share->auto_increment_table--;

#if MYSQL_VERSION_ID < 50500
    if (pthread_mutex_init(&share->mutex, MY_MUTEX_INIT_FAST))
#else
    if (mysql_mutex_init(vp_key_mutex_share,
      &share->mutex, MY_MUTEX_INIT_FAST))
#endif
    {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_init_mutex;
    }

#if MYSQL_VERSION_ID < 50500
    if (pthread_mutex_init(&share->init_mutex, MY_MUTEX_INIT_FAST))
#else
    if (mysql_mutex_init(vp_key_mutex_share_init,
      &share->init_mutex, MY_MUTEX_INIT_FAST))
#endif
    {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_init_init_mutex;
    }

    thr_lock_init(&share->lock);

#ifdef WITH_PARTITION_STORAGE_ENGINE
    if (!(share->partition_share =
      vp_get_pt_share(table, share, error_num)))
      goto error_get_pt_share;
#endif

    if (my_hash_insert(&vp_open_tables, (uchar*) share))
    {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_hash_insert;
    }
  }
  share->use_count++;
  pthread_mutex_unlock(&vp_tbl_mutex);

  vp->share = share;

  DBUG_PRINT("info",("vp share=%x", share));
  DBUG_RETURN(share);

error_hash_insert:
#ifdef WITH_PARTITION_STORAGE_ENGINE
error_get_pt_share:
#endif
  thr_lock_delete(&share->lock);
  pthread_mutex_destroy(&share->init_mutex);
error_init_init_mutex:
  pthread_mutex_destroy(&share->mutex);
error_init_mutex:
error_create_table_list:
error_parse_comment_string:
  vp_free_share_alloc(share);
  my_free(share, MYF(0));
error_alloc_share:
  pthread_mutex_unlock(&vp_tbl_mutex);
  DBUG_RETURN(NULL);
}

int vp_free_share(
  VP_SHARE *share
) {
  DBUG_ENTER("vp_free_share");
  pthread_mutex_lock(&vp_tbl_mutex);
  if (!--share->use_count)
  {
    vp_free_share_alloc(share);
    my_hash_delete(&vp_open_tables, (uchar*) share);
    thr_lock_delete(&share->lock);
    pthread_mutex_destroy(&share->init_mutex);
    pthread_mutex_destroy(&share->mutex);
    my_free(share, MYF(0));
  }
  pthread_mutex_unlock(&vp_tbl_mutex);
  DBUG_RETURN(0);
}

#ifdef WITH_PARTITION_STORAGE_ENGINE
VP_PARTITION_SHARE *vp_get_pt_share(
  TABLE *table,
  VP_SHARE *share,
  int *error_num
) {
  VP_PARTITION_SHARE *partition_share;
  char *tmp_name;
  DBUG_ENTER("vp_get_pt_share");

  pthread_mutex_lock(&vp_pt_share_mutex);
  if (!(partition_share = (VP_PARTITION_SHARE*) my_hash_search(
    &vp_open_pt_share,
    (uchar*) table->s->path.str, table->s->path.length)))
  {
    DBUG_PRINT("info",("vp create new pt share"));
    if (!(partition_share = (VP_PARTITION_SHARE *)
      my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
        &partition_share, sizeof(*partition_share),
        &tmp_name, table->s->path.length + 1,
        NullS))
    ) {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_alloc_share;
    }

    partition_share->use_count = 0;
    partition_share->table_name_length = table->s->path.length;
    partition_share->table_name = tmp_name;
    memcpy(partition_share->table_name, table->s->path.str,
      partition_share->table_name_length);

#if MYSQL_VERSION_ID < 50500
    if (pthread_mutex_init(&partition_share->pt_handler_mutex,
      MY_MUTEX_INIT_FAST))
#else
    if (mysql_mutex_init(vp_key_mutex_pt_handler,
      &partition_share->pt_handler_mutex, MY_MUTEX_INIT_FAST))
#endif
    {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_init_pt_handler_mutex;
    }

    if(
      my_hash_init(&partition_share->pt_handler_hash, system_charset_info,
        32, 0, 0, (my_hash_get_key) vp_pt_handler_share_get_key, 0, 0)
    ) {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_init_pt_handler_hash;
    }

    if (my_hash_insert(&vp_open_pt_share, (uchar*) partition_share))
    {
      *error_num = HA_ERR_OUT_OF_MEM;
      goto error_hash_insert;
    }
  }
  partition_share->use_count++;
  pthread_mutex_unlock(&vp_pt_share_mutex);

  DBUG_PRINT("info",("vp partition_share=%x", partition_share));
  DBUG_RETURN(partition_share);

error_hash_insert:
  my_hash_free(&partition_share->pt_handler_hash);
error_init_pt_handler_hash:
  pthread_mutex_destroy(&partition_share->pt_handler_mutex);
error_init_pt_handler_mutex:
  my_free(partition_share, MYF(0));
error_alloc_share:
  pthread_mutex_unlock(&vp_pt_share_mutex);
  DBUG_RETURN(NULL);
}

int vp_free_pt_share(
  VP_PARTITION_SHARE *partition_share
) {
  DBUG_ENTER("vp_free_pt_share");
  pthread_mutex_lock(&vp_pt_share_mutex);
  if (!--partition_share->use_count)
  {
    my_hash_delete(&vp_open_pt_share, (uchar*) partition_share);
    my_hash_free(&partition_share->pt_handler_hash);
    pthread_mutex_destroy(&partition_share->pt_handler_mutex);
    my_free(partition_share, MYF(0));
  }
  pthread_mutex_unlock(&vp_pt_share_mutex);
  DBUG_RETURN(0);
}
#endif

bool vp_flush_logs(
  handlerton *hton
) {
  DBUG_ENTER("vp_flush_logs");
  DBUG_RETURN(FALSE);
}

handler* vp_create_handler(
  handlerton *hton,
  TABLE_SHARE *table, 
  MEM_ROOT *mem_root
) {
  DBUG_ENTER("vp_create_handler");
  DBUG_RETURN(new (mem_root) ha_vp(hton, table));
}

int vp_close_connection(
  handlerton*	hton,
  THD* thd
) {
  DBUG_ENTER("vp_close_connection");
  DBUG_RETURN(0);
}

void vp_drop_database(
  handlerton *hton,
  char* path
) {
  DBUG_ENTER("vp_drop_database");
  DBUG_VOID_RETURN;
}

bool vp_show_status(
  handlerton *hton,
  THD *thd, 
  stat_print_fn *stat_print,
  enum ha_stat_type stat_type
) {
  DBUG_ENTER("vp_show_status");
  switch (stat_type) {
    case HA_ENGINE_STATUS:
    default:
      DBUG_RETURN(FALSE);
	}
}

int vp_start_consistent_snapshot(
  handlerton *hton,
  THD* thd
) {
  DBUG_ENTER("vp_start_consistent_snapshot");
  DBUG_RETURN(0);
}

int vp_commit(
  handlerton *hton,
  THD *thd,
  bool all
) {
  DBUG_ENTER("vp_commit");
  DBUG_RETURN(0);
}

int vp_rollback(
  handlerton *hton,
  THD *thd,
  bool all
) {
  DBUG_ENTER("vp_rollback");
  DBUG_RETURN(0);
}

int vp_xa_prepare(
  handlerton *hton,
  THD* thd,
  bool all
) {
  DBUG_ENTER("vp_xa_prepare");
  DBUG_RETURN(0);
}

int vp_xa_recover(
  handlerton *hton,
  XID* xid_list,
  uint len
) {
  DBUG_ENTER("vp_xa_recover");
  DBUG_RETURN(0);
}

int vp_xa_commit_by_xid(
  handlerton *hton,
  XID* xid
) {
  DBUG_ENTER("vp_xa_commit_by_xid");
  DBUG_RETURN(0);
}

int vp_xa_rollback_by_xid(
  handlerton *hton,
  XID* xid
) {
  DBUG_ENTER("vp_xa_rollback_by_xid");
  DBUG_RETURN(0);
}

int vp_db_done(
  void *p
) {
  DBUG_ENTER("vp_db_done");
#ifdef WITH_PARTITION_STORAGE_ENGINE
  my_hash_free(&vp_open_pt_share);
#endif
  my_hash_free(&vp_open_tables);
#ifdef WITH_PARTITION_STORAGE_ENGINE
  pthread_mutex_destroy(&vp_pt_share_mutex);
#endif
  pthread_mutex_destroy(&vp_tbl_mutex);
#ifndef WITHOUT_VP_BG_ACCESS
  pthread_attr_destroy(&vp_pt_attr);
#endif
  DBUG_RETURN(0);
}

int vp_panic(
  handlerton *hton,
  ha_panic_function type
) {
  DBUG_ENTER("vp_panic");
  DBUG_RETURN(0);
}

int vp_db_init(
  void *p
) {
  int error_num;
  handlerton *vp_hton = (handlerton *)p;
  DBUG_ENTER("vp_db_init");
  vp_hton_ptr = vp_hton;

  vp_hton->state = SHOW_OPTION_YES;
#ifdef PARTITION_HAS_EXTRA_ATTACH_CHILDREN
  vp_hton->flags = HTON_NO_FLAGS;
#else
  vp_hton->flags = HTON_NO_PARTITION;
#endif
#ifdef HTON_CAN_MULTISTEP_MERGE
  vp_hton->flags |= HTON_CAN_MULTISTEP_MERGE;
#endif
  /* vp_hton->db_type = DB_TYPE_VP; */
  /*
  vp_hton->savepoint_offset;
  vp_hton->savepoint_set = vp_savepoint_set;
  vp_hton->savepoint_rollback = vp_savepoint_rollback;
  vp_hton->savepoint_release = vp_savepoint_release;
  vp_hton->create_cursor_read_view = vp_create_cursor_read_view;
  vp_hton->set_cursor_read_view = vp_set_cursor_read_view;
  vp_hton->close_cursor_read_view = vp_close_cursor_read_view;
  */
  vp_hton->panic = vp_panic;
  vp_hton->close_connection = vp_close_connection;
  vp_hton->start_consistent_snapshot = vp_start_consistent_snapshot;
  vp_hton->flush_logs = vp_flush_logs;
  vp_hton->commit = vp_commit;
  vp_hton->rollback = vp_rollback;
  if (vp_support_xa)
  {
    vp_hton->prepare = vp_xa_prepare;
    vp_hton->recover = vp_xa_recover;
    vp_hton->commit_by_xid = vp_xa_commit_by_xid;
    vp_hton->rollback_by_xid = vp_xa_rollback_by_xid;
  }
  vp_hton->create = vp_create_handler;
  vp_hton->drop_database = vp_drop_database;
  vp_hton->show_status = vp_show_status;

#ifdef HAVE_PSI_INTERFACE
  init_vp_psi_keys();
#endif

#ifndef WITHOUT_VP_BG_ACCESS
  if (pthread_attr_init(&vp_pt_attr))
  {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_pt_attr_init;
  }
  if (pthread_attr_setdetachstate(&vp_pt_attr, PTHREAD_CREATE_DETACHED))
  {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_pt_attr_setstate;
  }
#endif

#if MYSQL_VERSION_ID < 50500
  if (pthread_mutex_init(&vp_tbl_mutex, MY_MUTEX_INIT_FAST))
#else
  if (mysql_mutex_init(vp_key_mutex_tbl,
    &vp_tbl_mutex, MY_MUTEX_INIT_FAST))
#endif
  {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_tbl_mutex_init;
  }
#ifdef WITH_PARTITION_STORAGE_ENGINE
#if MYSQL_VERSION_ID < 50500
  if (pthread_mutex_init(&vp_pt_share_mutex, MY_MUTEX_INIT_FAST))
#else
  if (mysql_mutex_init(vp_key_mutex_pt_share,
    &vp_pt_share_mutex, MY_MUTEX_INIT_FAST))
#endif
  {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_pt_share_mutex_init;
  }
#endif

  if(
    my_hash_init(&vp_open_tables, system_charset_info, 32, 0, 0,
                   (my_hash_get_key) vp_tbl_get_key, 0, 0)
  ) {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_open_tables_hash_init;
  }
#ifdef WITH_PARTITION_STORAGE_ENGINE
  if(
    my_hash_init(&vp_open_pt_share, system_charset_info, 32, 0, 0,
                   (my_hash_get_key) vp_pt_share_get_key, 0, 0)
  ) {
    error_num = HA_ERR_OUT_OF_MEM;
    goto error_open_pt_share_hash_init;
  }
#endif

  DBUG_RETURN(0);

#ifdef WITH_PARTITION_STORAGE_ENGINE
error_open_pt_share_hash_init:
#endif
  my_hash_free(&vp_open_tables);
error_open_tables_hash_init:
#ifdef WITH_PARTITION_STORAGE_ENGINE
  pthread_mutex_destroy(&vp_pt_share_mutex);
error_pt_share_mutex_init:
#endif
  pthread_mutex_destroy(&vp_tbl_mutex);
error_tbl_mutex_init:
#ifndef WITHOUT_VP_BG_ACCESS
error_pt_attr_setstate:
  pthread_attr_destroy(&vp_pt_attr);
error_pt_attr_init:
#endif
  DBUG_RETURN(error_num);
}

char *vp_create_string(
  const char *str,
  uint length
) {
  char *res;
  DBUG_ENTER("vp_create_string");
  if (!(res = (char*) my_malloc(length + 1, MYF(MY_WME))))
    DBUG_RETURN(NULL);
  memcpy(res, str, length);
  res[length] = '\0';
  DBUG_RETURN(res);
}

char *vp_create_table_name_string(
  const char *table_name,
  const char *part_name,
  const char *sub_name
) {
  char *res, *tmp;
  uint length = strlen(table_name);
  DBUG_ENTER("vp_create_table_name_string");
  if (part_name)
  {
    length += sizeof("#P#") - 1 + strlen(part_name);
    if (sub_name)
      length += sizeof("#SP#") - 1 + strlen(sub_name);
  }
  if (!(res = (char*) my_malloc(length + 1, MYF(MY_WME))))
    DBUG_RETURN(NULL);
  tmp = strmov(res, table_name);
  if (part_name)
  {
    tmp = strmov(tmp, "#P#");
    tmp = strmov(tmp, part_name);
    if (sub_name)
    {
      tmp = strmov(tmp, "#SP#");
      tmp = strmov(tmp, sub_name);
    }
  }
  DBUG_RETURN(res);
}

#ifdef WITH_PARTITION_STORAGE_ENGINE
void vp_get_partition_info(
  const char *table_name,
  const TABLE *table,
  partition_element **part_elem,
  partition_element **sub_elem
) {
  char tmp_name[FN_LEN];
  partition_info *part_info = table->part_info;
  DBUG_ENTER("vp_get_partition_info");
  *part_elem = NULL;
  *sub_elem = NULL;
  if (!part_info)
    DBUG_VOID_RETURN;

  DBUG_PRINT("info",("vp table_name=%s", table_name));
  List_iterator<partition_element> part_it(part_info->partitions);
  while ((*part_elem = part_it++))
  {
    if ((*part_elem)->subpartitions.elements)
    {
      List_iterator<partition_element> sub_it((*part_elem)->subpartitions);
      while ((*sub_elem = sub_it++))
      {
        create_subpartition_name(tmp_name, table->s->path.str,
          (*part_elem)->partition_name, (*sub_elem)->partition_name,
          NORMAL_PART_NAME);
        DBUG_PRINT("info",("vp tmp_name=%s", tmp_name));
        if (!strcmp(table_name, tmp_name))
          DBUG_VOID_RETURN;
      }
    } else {
      create_partition_name(tmp_name, table->s->path.str,
        (*part_elem)->partition_name, NORMAL_PART_NAME, TRUE);
      DBUG_PRINT("info",("vp tmp_name=%s", tmp_name));
      if (!strcmp(table_name, tmp_name))
        DBUG_VOID_RETURN;
    }
  }
  *part_elem = NULL;
  *sub_elem = NULL;
  DBUG_PRINT("info",("vp no hit"));
  DBUG_VOID_RETURN;
}
#endif

int vp_create_table_list(
  VP_SHARE *share
) {
  int table_count, roop_count, length;
  char *tmp_ptr, *tmp_ptr2, *tmp_ptr3, *tmp_name_ptr, *tmp_path_ptr;
  DBUG_ENTER("vp_create_table_list");
  table_count = 1;
  tmp_ptr = share->tgt_table_name_list;
  while (*tmp_ptr == ' ')
    tmp_ptr++;
  while (TRUE)
  {
    if ((tmp_ptr2 = strchr(tmp_ptr, ' ')))
    {
      table_count++;
      tmp_ptr = tmp_ptr2 + 1;
      while (*tmp_ptr == ' ')
        tmp_ptr++;
    } else
      break;
  }
  share->use_tables_size = (table_count + 7) / 8;
  if (!(share->tgt_db_name = (char**)
    my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
      &share->tgt_db_name, sizeof(char*) * table_count,
      &share->tgt_table_name, sizeof(char*) * table_count,
      &share->part_tables, sizeof(TABLE_LIST) * table_count,
      &tmp_name_ptr, sizeof(char) * (
        share->tgt_table_name_list_length +
        share->tgt_default_db_name_length * table_count +
        share->tgt_table_name_prefix_length * table_count +
        share->tgt_table_name_suffix_length * table_count +
        2 * table_count
      ),
      &tmp_path_ptr, sizeof(char) * (
        share->tgt_table_name_list_length +
        share->tgt_default_db_name_length * table_count +
        share->tgt_table_name_prefix_length * table_count +
        share->tgt_table_name_suffix_length * table_count +
        4 * table_count
      ),
      &share->select_ignore,
      sizeof(uchar) * share->use_tables_size,
      &share->select_ignore_with_lock,
      sizeof(uchar) * share->use_tables_size,
      NullS))
  )
    DBUG_RETURN(HA_ERR_OUT_OF_MEM);

  tmp_ptr = share->tgt_table_name_list;
  while (*tmp_ptr == ' ')
    tmp_ptr++;
  roop_count = 0;
  while (TRUE)
  {
    if ((tmp_ptr2 = strchr(tmp_ptr, ' ')))
      *tmp_ptr2 = '\0';

    share->tgt_db_name[roop_count] =
      share->part_tables[roop_count].db = tmp_name_ptr;

    if ((tmp_ptr3 = strchr(tmp_ptr, '.')))
    {
      /* exist database name */
      *tmp_ptr3 = '\0';
      length = strlen(tmp_ptr);
      memcpy(tmp_name_ptr, tmp_ptr, length + 1);
      tmp_name_ptr += length + 1;
      tmp_ptr = tmp_ptr3 + 1;
      share->part_tables[roop_count].db_length = length;
    } else {
      memcpy(tmp_name_ptr, share->tgt_default_db_name,
        share->tgt_default_db_name_length + 1);
      tmp_name_ptr += share->tgt_default_db_name_length + 1;
      share->part_tables[roop_count].db_length =
        share->tgt_default_db_name_length;
    }

    share->tgt_table_name[roop_count] =
      share->part_tables[roop_count].alias =
      share->part_tables[roop_count].table_name = tmp_name_ptr;
    memcpy(tmp_name_ptr, share->tgt_table_name_prefix,
      share->tgt_table_name_prefix_length);
    tmp_name_ptr += share->tgt_table_name_prefix_length;

    length = strlen(tmp_ptr);
    memcpy(tmp_name_ptr, tmp_ptr, length);
    tmp_name_ptr += length;

    memcpy(tmp_name_ptr, share->tgt_table_name_suffix,
      share->tgt_table_name_suffix_length + 1);
    tmp_name_ptr += share->tgt_table_name_suffix_length + 1;

    share->part_tables[roop_count].table_name_length =
      share->tgt_table_name_prefix_length + length +
      share->tgt_table_name_suffix_length;

    DBUG_PRINT("info",("vp db=%s",
      share->part_tables[roop_count].db));
    DBUG_PRINT("info",("vp table_name=%s",
      share->part_tables[roop_count].table_name));

    if (!tmp_ptr2)
      break;
    tmp_ptr = tmp_ptr2 + 1;
    while (*tmp_ptr == ' ')
      tmp_ptr++;
    roop_count++;
  }
  share->table_count = table_count;
  DBUG_RETURN(0);
}

int vp_correspond_columns(
  TABLE *table,
  VP_SHARE *share,
  TABLE_SHARE *table_share,
  TABLE_LIST *part_tables
) {
  int error_num, roop_count, roop_count2, roop_count3, field_count, field_idx,
    key_count, bitmap_size, max_count, min_count;
  int *correspond_columns_p, *tmp_correspond_columns_p;
  int **correspond_columns_c_ptr, *correspond_columns_c,
    *tmp_correspond_columns_c;
  uchar *correspond_columns_bit, *tmp_correspond_columns_bit;
  VP_KEY *keys, *largest_key, *tmp_keys, *tmp_keys2, *tmp_keys3;
  uchar *keys_bit, *tmp_keys_bit;
  VP_CORRESPOND_KEY **correspond_pk, *correspond_keys_p,
    *tmp_correspond_keys_p, *tmp_correspond_keys_p2,
    *tmp_correspond_keys_p3, *tmp_correspond_keys_p4, *tmp_correspond_keys_p5,
    **correspond_keys_p_ptr;
  uchar *correspond_keys_bit, *tmp_correspond_keys_bit;
  bool correspond_flag, tmp_correspond_flag;
  uchar *need_converting, *need_searching, *need_full_col_for_update,
    *pk_in_read_index, *cpy_clm_bitmap;
  TABLE *part_table;
  TABLE_SHARE *part_table_share;
  Field *field, *field2;
  KEY *key_info, *key_info2;
  KEY_PART_INFO *key_part, *key_part2, *key_part3;
  handlerton *hton;
  ulong *def_versions;
  uint key_parts2;
  DBUG_ENTER("vp_correspond_columns");
  pthread_mutex_lock(&share->init_mutex);
  if (share->init && share->reinit)
  {
    correspond_flag = TRUE;
    def_versions = share->def_versions;
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      if (def_versions[roop_count] !=
        part_tables[roop_count].table->s->get_table_def_version())
      {
        correspond_flag = FALSE;
        break;
      }
    }
    if (correspond_flag)
      share->reinit = FALSE;
    else
      my_free(share->correspond_columns_p, MYF(0));
  }

  if (!share->init || share->reinit)
  {
    field_count = 0;
    key_count = 0;
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      part_table_share = part_tables[roop_count].table->s;
      field_count += part_table_share->fields;
      key_count += part_table_share->keys;
    }

    bitmap_size = (table_share->fields + 7) / 8;
    if (!(correspond_columns_p = (int *)
      my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
        &correspond_columns_p,
        sizeof(int) * table_share->fields * share->table_count,
        &correspond_columns_c_ptr,
        sizeof(int *) * share->table_count,
        &correspond_columns_c,
        sizeof(int) * field_count,
        &correspond_columns_bit,
        sizeof(uchar) * bitmap_size * (share->table_count + 1),
        &keys,
        sizeof(VP_KEY) * table_share->keys,
        &keys_bit,
        sizeof(uchar) * bitmap_size * table_share->keys,
        &correspond_pk,
        sizeof(VP_CORRESPOND_KEY *) * share->table_count,
        &correspond_keys_p_ptr,
        sizeof(VP_CORRESPOND_KEY *) * share->table_count,
        &correspond_keys_p,
        sizeof(VP_CORRESPOND_KEY) * key_count,
        &correspond_keys_bit,
        sizeof(uchar) * bitmap_size * key_count,
        &cpy_clm_bitmap, sizeof(uchar) * bitmap_size,
        &need_converting,
        sizeof(uchar) * share->use_tables_size,
        &need_searching,
        sizeof(uchar) * share->use_tables_size,
        &need_full_col_for_update,
        sizeof(uchar) * share->use_tables_size,
        &pk_in_read_index,
        sizeof(uchar) * share->use_tables_size,
        &def_versions,
        sizeof(ulong) * share->table_count,
        NullS))
    ) {
      error_num = HA_ERR_OUT_OF_MEM;
      goto error_alloc;
    }

    tmp_correspond_columns_p = correspond_columns_p;
    for (roop_count = 0; roop_count < share->table_count;
      roop_count++, tmp_correspond_columns_p++)
    {
      for (roop_count2 = 0; roop_count2 < table_share->fields;
        roop_count2++, tmp_correspond_columns_p++)
        *tmp_correspond_columns_p = MAX_FIELDS;
    }

    tmp_correspond_columns_p = correspond_columns_p;
    tmp_correspond_columns_c = correspond_columns_c;
    tmp_correspond_columns_bit =
      correspond_columns_bit + bitmap_size;
    tmp_correspond_keys_p = correspond_keys_p;
    tmp_correspond_keys_p2 = NULL;
    tmp_correspond_keys_bit = correspond_keys_bit;
    for (roop_count = 0; roop_count < share->table_count; roop_count++)
    {
      correspond_columns_c_ptr[roop_count] = tmp_correspond_columns_c;
      correspond_keys_p_ptr[roop_count] = tmp_correspond_keys_p;
      part_table = part_tables[roop_count].table;
      part_table_share = part_table->s;
      hton = part_table_share->db_type();
      if (hton->db_type == DB_TYPE_PARTITION_DB)
        hton = part_table_share->default_part_db_type;
      if (
        hton->db_type == DB_TYPE_HEAP ||
        hton->db_type == DB_TYPE_MYISAM ||
        hton->db_type == DB_TYPE_MRG_MYISAM
      )
        vp_set_bit(need_full_col_for_update, roop_count);

      if (
        part_table->file->ha_table_flags() & HA_PRIMARY_KEY_IN_READ_INDEX &&
        part_table_share->primary_key != MAX_KEY
      )
        vp_set_bit(pk_in_read_index, roop_count);

      correspond_flag = FALSE;
      DBUG_PRINT("info",("vp correspond table %s",
        part_table_share->table_name.str));
      for (roop_count2 = 0; roop_count2 < part_table_share->fields;
        roop_count2++)
      {
        if ((field = find_field_in_table_sef(table,
          part_table_share->fieldnames.type_names[roop_count2])))
        {
          DBUG_PRINT("info",("vp correspond column %s",
            part_table_share->fieldnames.type_names[roop_count2]));

          field2 = part_table->field[roop_count2];
          if (field->type() != field2->type())
          {
            error_num = ER_VP_DIFFERENT_COLUMN_TYPE_NUM;
            my_printf_error(error_num, ER_VP_DIFFERENT_COLUMN_TYPE_STR, MYF(0),
              part_table_share->table_name.str,
              part_table_share->fieldnames.type_names[roop_count2]);
            goto error_table_correspond;
          }
          if (
            field->charset() != field2->charset() ||
            ((field->null_bit == 0) != (field2->null_bit == 0)) ||
            field->pack_length() != field2->pack_length()
          ) {
            DBUG_PRINT("info",("vp need converting"));
            DBUG_PRINT("info",("vp charset %x %x",
              field->charset(), field2->charset()));
            DBUG_PRINT("info",("vp null_bit %d %d",
              field->null_bit, field2->null_bit));
            DBUG_PRINT("info",("vp pack_length %ld %ld",
              field->pack_length(), field2->pack_length()));
            vp_set_bit(need_converting, roop_count);
          }

          field_idx = field->field_index;
          tmp_correspond_columns_p[field_idx] = roop_count2;
          tmp_correspond_columns_c[roop_count2] = field_idx;
          vp_set_bit(tmp_correspond_columns_bit, field_idx);
          vp_set_bit(correspond_columns_bit, field_idx);
          correspond_flag = TRUE;
        } else {
          DBUG_PRINT("info",("vp uncorrespond column %s",
            part_table_share->fieldnames.type_names[roop_count2]));
          tmp_correspond_columns_c[roop_count2] = MAX_FIELDS;
        }
      }
      if (!correspond_flag)
      {
        error_num = ER_VP_CANT_CORRESPOND_TABLE_NUM;
        my_printf_error(error_num, ER_VP_CANT_CORRESPOND_TABLE_STR, MYF(0),
          part_table_share->table_name.str);
        goto error_table_correspond;
      }

      part_table_share = part_tables[roop_count].table->s;
      key_info = part_table_share->key_info;
      for (roop_count2 = 0; roop_count2 < part_table_share->keys;
        roop_count2++)
      {
        tmp_correspond_keys_p->table_idx = roop_count;
        tmp_correspond_keys_p->key_idx = roop_count2;
        tmp_correspond_keys_p->columns_bit = tmp_correspond_keys_bit;
        tmp_correspond_keys_p->next = tmp_correspond_keys_p2;
        tmp_correspond_keys_p2 = tmp_correspond_keys_p;
        for (
          roop_count3 = 0, key_part = key_info[roop_count2].key_part;
          roop_count3 < key_info[roop_count2].key_parts;
          roop_count3++, key_part++
        ) {
          field_idx = tmp_correspond_columns_c[key_part->field->field_index];
          if (field_idx < MAX_FIELDS)
            vp_set_bit(tmp_correspond_keys_bit, field_idx);
        }
        if (vp_bit_is_set(pk_in_read_index, roop_count))
        {
          for (
            roop_count3 = 0,
            key_part = key_info[part_table_share->primary_key].key_part;
            roop_count3 < key_info[part_table_share->primary_key].key_parts;
            roop_count3++, key_part++
          ) {
            field_idx = tmp_correspond_columns_c[key_part->field->field_index];
            if (field_idx < MAX_FIELDS)
              vp_set_bit(tmp_correspond_keys_bit, field_idx);
          }
        }
        tmp_correspond_keys_p++;
        tmp_correspond_keys_bit += bitmap_size;
      }

      tmp_correspond_columns_p += table_share->fields;
      tmp_correspond_columns_c += part_table_share->fields;
      tmp_correspond_columns_bit += bitmap_size;
    }
    tmp_correspond_keys_p4 = tmp_correspond_keys_p2;

    for (roop_count = 0; roop_count < table_share->fields; roop_count++)
    {
      if (!(vp_bit_is_set(correspond_columns_bit, roop_count)))
      {
        error_num = ER_VP_CANT_CORRESPOND_COLUMN_NUM;
        my_printf_error(error_num, ER_VP_CANT_CORRESPOND_COLUMN_STR, MYF(0),
          table_share->fieldnames.type_names[roop_count]);
        goto error_column_correspond;
      }
    }

    tmp_keys_bit = keys_bit;
    key_info = table_share->key_info;
    for (roop_count = 0; roop_count < table_share->keys;
      roop_count++)
    {
      keys[roop_count].key_idx = roop_count;
      keys[roop_count].columns_bit = tmp_keys_bit;
      if (roop_count < table_share->keys - 1)
      {
        keys[roop_count].key_length_next = &keys[roop_count + 1];
      }
      if (roop_count > 0)
      {
        keys[roop_count].key_length_prev = &keys[roop_count - 1];
      }
      for (
        roop_count2 = 0, key_part = key_info[roop_count].key_part;
        roop_count2 < key_info[roop_count].key_parts;
        roop_count2++, key_part++
      ) {
        field_idx = key_part->field->field_index;
        vp_set_bit(tmp_keys_bit, field_idx);
      }
      tmp_keys_bit += bitmap_size;
    }

    tmp_keys = largest_key = &keys[0];
    while (tmp_keys)
    {
      max_count = key_info[tmp_keys->key_idx].key_parts;
      tmp_keys2 = tmp_keys->key_length_next;
      tmp_keys3 = tmp_keys;
      while (tmp_keys2)
      {
        if (max_count < key_info[tmp_keys2->key_idx].key_parts)
        {
          max_count = key_info[tmp_keys2->key_idx].key_parts;
          tmp_keys = tmp_keys2;
        }
        tmp_keys2 = tmp_keys2->key_length_next;
      }
      if (tmp_keys != tmp_keys3)
      {
        if (tmp_keys->key_length_next)
          tmp_keys->key_length_next->key_length_prev =
            tmp_keys->key_length_prev;
        tmp_keys->key_length_prev->key_length_next =
          tmp_keys->key_length_next;
        tmp_keys->key_length_next = tmp_keys3;
        tmp_keys->key_length_prev = tmp_keys3->key_length_prev;
        if (tmp_keys3->key_length_prev)
          tmp_keys3->key_length_prev->key_length_next = tmp_keys;
        else
          largest_key = tmp_keys;
        tmp_keys3->key_length_prev = tmp_keys;
      }
      DBUG_PRINT("info",("vp key_length=%d key_id=%d",
        key_info[tmp_keys->key_idx].key_parts, tmp_keys->key_idx));
      tmp_keys = tmp_keys->key_length_next;
    }

    /* correspond keys */
    tmp_keys = largest_key;
    tmp_correspond_keys_p = tmp_correspond_keys_p4;
    while (tmp_keys)
    {
      correspond_flag = FALSE;
      min_count = MAX_FIELDS;
      tmp_correspond_keys_p2 = tmp_correspond_keys_p;
      tmp_correspond_keys_p3 = NULL;
      while (tmp_correspond_keys_p2)
      {
        tmp_correspond_flag = TRUE;
        for (roop_count = 0; roop_count < bitmap_size; roop_count++)
        {
          if ((tmp_keys->columns_bit[roop_count] &
            tmp_correspond_keys_p2->columns_bit[roop_count]) !=
            tmp_keys->columns_bit[roop_count])
          {
            DBUG_PRINT("info",("vp uncorrespond key p1 %d and child %d-%d",
              tmp_keys->key_idx, tmp_correspond_keys_p2->table_idx,
              tmp_correspond_keys_p2->key_idx));
            tmp_correspond_flag = FALSE;
            break;
          }
        }
        if (tmp_correspond_flag)
        {
          part_table_share =
            part_tables[tmp_correspond_keys_p2->table_idx].table->s;
          key_info2 = part_table_share->key_info;
          if (
            (key_info[tmp_keys->key_idx].flags & HA_FULLTEXT) !=
            (key_info2[tmp_correspond_keys_p2->key_idx].flags & HA_FULLTEXT) ||
            (key_info[tmp_keys->key_idx].flags & HA_SPATIAL) !=
            (key_info2[tmp_correspond_keys_p2->key_idx].flags & HA_SPATIAL) ||
            key_info[tmp_keys->key_idx].algorithm !=
              key_info2[tmp_correspond_keys_p2->key_idx].algorithm
          ) {
            DBUG_PRINT("info",("vp uncorrespond key p2 %d and child %d-%d",
              tmp_keys->key_idx, tmp_correspond_keys_p2->table_idx,
              tmp_correspond_keys_p2->key_idx));
            DBUG_PRINT("info",("vp flags=%u %u",
              key_info[tmp_keys->key_idx].flags,
              key_info2[tmp_correspond_keys_p2->key_idx].flags));
            DBUG_PRINT("info",("vp algorithm=%u %u",
              key_info[tmp_keys->key_idx].algorithm,
              key_info2[tmp_correspond_keys_p2->key_idx].algorithm));
            tmp_correspond_flag = FALSE;
          } else if (
            (key_info[tmp_keys->key_idx].flags & HA_FULLTEXT) &&
            key_info[tmp_keys->key_idx].key_parts !=
              key_info2[tmp_correspond_keys_p2->key_idx].key_parts
          ) {
            DBUG_PRINT("info",("vp uncorrespond key p2 %d and child %d-%d",
              tmp_keys->key_idx, tmp_correspond_keys_p2->table_idx,
              tmp_correspond_keys_p2->key_idx));
            DBUG_PRINT("info",("vp key_parts=%u %u",
              key_info[tmp_keys->key_idx].key_parts,
              key_info2[tmp_correspond_keys_p2->key_idx].key_parts));
            tmp_correspond_flag = FALSE;
          }
        }
        if (tmp_correspond_flag)
        {
          key_part = key_info[tmp_keys->key_idx].key_part;
          key_part2 = key_info2[tmp_correspond_keys_p2->key_idx].key_part;
          key_parts2 = key_info2[tmp_correspond_keys_p2->key_idx].key_parts;
          if (vp_bit_is_set(pk_in_read_index,
            tmp_correspond_keys_p2->table_idx))
            key_part3 = key_info2[part_table_share->primary_key].key_part;
          for (roop_count = 0;
            roop_count < key_info[tmp_keys->key_idx].key_parts; roop_count++)
          {
            field = key_part[roop_count].field;
            if (roop_count < key_parts2)
              field2 = key_part2[roop_count].field;
            else
              field2 = key_part3[roop_count - key_parts2].field;
            if (
              correspond_columns_p[table_share->fields *
              tmp_correspond_keys_p2->table_idx + field->field_index] !=
              field2->field_index
            ) {
              DBUG_PRINT("info",("vp uncorrespond key p3 %d and child %d-%d",
                tmp_keys->key_idx, tmp_correspond_keys_p2->table_idx,
                tmp_correspond_keys_p2->key_idx));
              tmp_correspond_flag = FALSE;
              break;
            }
          }
        }
        if (tmp_correspond_flag)
        {
          DBUG_PRINT("info",("vp correspond key %d and child %d-%d",
            tmp_keys->key_idx, tmp_correspond_keys_p2->table_idx,
            tmp_correspond_keys_p2->key_idx));
          if (tmp_correspond_keys_p2 == tmp_correspond_keys_p)
            tmp_correspond_keys_p = tmp_correspond_keys_p->next;
          else
            tmp_correspond_keys_p3->next = tmp_correspond_keys_p2->next;
          tmp_correspond_keys_p4 = tmp_correspond_keys_p2->next;
          tmp_correspond_keys_p2->next = tmp_keys->correspond_key;
          tmp_keys->correspond_key = tmp_correspond_keys_p2;
          correspond_flag = TRUE;
          tmp_correspond_keys_p2->key_parts =
            key_info2[tmp_correspond_keys_p2->key_idx].key_parts;
          if (min_count > tmp_correspond_keys_p2->key_parts)
          {
            min_count = tmp_correspond_keys_p2->key_parts;
            tmp_correspond_keys_p2->next_shortest =
              tmp_keys->shortest_correspond_key;
            tmp_keys->shortest_correspond_key = tmp_correspond_keys_p2;
          } else {
            tmp_correspond_keys_p5 = tmp_keys->shortest_correspond_key;
            while (TRUE)
            {
              if (!tmp_correspond_keys_p5->next_shortest)
              {
                tmp_correspond_keys_p5->next_shortest = tmp_correspond_keys_p2;
                break;
              }
              if (tmp_correspond_keys_p5->next_shortest->key_parts >
                tmp_correspond_keys_p2->key_parts)
              {
                tmp_correspond_keys_p2->next_shortest =
                  tmp_correspond_keys_p5->next_shortest;
                tmp_correspond_keys_p5->next_shortest = tmp_correspond_keys_p2;
                break;
              }
              tmp_correspond_keys_p5 = tmp_correspond_keys_p5->next_shortest;
            }
          }
          tmp_correspond_keys_p2 = tmp_correspond_keys_p4;
        } else {
          tmp_correspond_keys_p3 = tmp_correspond_keys_p2;
          tmp_correspond_keys_p2 = tmp_correspond_keys_p2->next;
        }
      }

      if (!correspond_flag && (tmp_keys2 = tmp_keys->key_length_prev))
      {
        DBUG_PRINT("info",("vp check key with same columns"));
        while (key_info[tmp_keys->key_idx].key_parts ==
          key_info[tmp_keys2->key_idx].key_parts)
        {
          tmp_correspond_flag = TRUE;
          for (roop_count = 0; roop_count < bitmap_size; roop_count++)
          {
            if (tmp_keys->columns_bit[roop_count] !=
              tmp_keys2->columns_bit[roop_count])
            {
              DBUG_PRINT("info",("vp uncorrespond key(same size) %d and %d",
                tmp_keys->key_idx, tmp_keys2->key_idx));
              tmp_correspond_flag = FALSE;
              break;
            }
          }
          if (tmp_correspond_flag)
          {
            DBUG_PRINT("info",("vp correspond key(same size) %d and %d",
              tmp_keys->key_idx, tmp_keys2->key_idx));
            tmp_keys->correspond_key = tmp_keys2->correspond_key;
            tmp_keys->shortest_correspond_key =
              tmp_keys2->shortest_correspond_key;
            correspond_flag = TRUE;
            break;
          }
          if (!(tmp_keys2 = tmp_keys2->key_length_prev))
            break;
        }
      }

      if (!correspond_flag)
      {
        error_num = ER_VP_CANT_CORRESPOND_KEY_NUM;
        my_printf_error(error_num, ER_VP_CANT_CORRESPOND_KEY_STR, MYF(0),
          tmp_keys->key_idx);
        goto error_key_correspond;
      }
      tmp_keys = tmp_keys->key_length_next;
    }

    /* PK check */
    DBUG_PRINT("info",("vp PK check start"));
    if (share->pk_correspond_mode == 0)
    {
      tmp_correspond_keys_p = correspond_keys_p;
      key_info = &table_share->key_info[table_share->primary_key];
      key_part = key_info->key_part;
      for (roop_count = 0; roop_count < share->table_count; roop_count++)
      {
        DBUG_PRINT("info",("vp roop_count=%d", roop_count));
        part_table_share = part_tables[roop_count].table->s;
        DBUG_PRINT("info",("vp part_table_share->primary_key=%d",
          part_table_share->primary_key));
        DBUG_PRINT("info",("vp parent key_parts=%d",
          key_info->key_parts));
        DBUG_PRINT("info",("vp child key_parts=%d",
          part_table_share->key_info[part_table_share->primary_key].
          key_parts));
        if (
          part_table_share->primary_key == MAX_KEY ||
          key_info->key_parts !=
          part_table_share->key_info[part_table_share->primary_key].key_parts
        ) {
          error_num = ER_VP_CANT_CORRESPOND_PK_NUM;
          my_printf_error(error_num, ER_VP_CANT_CORRESPOND_PK_STR, MYF(0),
            part_table_share->table_name.str);
          goto error_key_correspond;
        }
        key_info2 = &part_table_share->key_info[part_table_share->primary_key];
        key_part2 = key_info2->key_part;
        for (roop_count2 = 0; roop_count2 < key_info->key_parts; roop_count2++)
        {
          DBUG_PRINT("info",("vp roop_count2=%d", roop_count2));
          field = key_part[roop_count2].field;
          field2 = key_part2[roop_count2].field;
          DBUG_PRINT("info",("vp correspond field index=%d",
            correspond_columns_p[table_share->fields *
            roop_count + field->field_index]));
          DBUG_PRINT("info",("vp current field index=%d",
            field2->field_index));
          if (
            correspond_columns_p[table_share->fields *
            roop_count + field->field_index] !=
            field2->field_index
          ) {
            error_num = ER_VP_CANT_CORRESPOND_PK_NUM;
            my_printf_error(error_num, ER_VP_CANT_CORRESPOND_PK_STR, MYF(0),
              part_table_share->table_name.str);
            goto error_key_correspond;
          }
        }
        correspond_pk[roop_count] =
          &tmp_correspond_keys_p[part_table_share->primary_key];
        tmp_correspond_keys_p += part_table_share->keys;
      }
    } else {
      tmp_correspond_keys_p = keys[table_share->primary_key].correspond_key;
      for (roop_count = 0; roop_count < share->table_count; roop_count++)
      {
        if (
          !tmp_correspond_keys_p ||
          tmp_correspond_keys_p->table_idx != roop_count
        ) {
          part_table_share = part_tables[roop_count].table->s;
          error_num = ER_VP_CANT_CORRESPOND_PK_NUM;
          my_printf_error(error_num, ER_VP_CANT_CORRESPOND_PK_STR, MYF(0),
            part_table_share->table_name.str);
          goto error_key_correspond;
        }

        key_info2 = part_tables[roop_count].table->s->key_info;
        min_count = key_info2[tmp_correspond_keys_p->key_idx].key_parts;
        correspond_pk[roop_count] = tmp_correspond_keys_p;
        tmp_correspond_keys_p = tmp_correspond_keys_p->next;
        while (
          tmp_correspond_keys_p &&
          tmp_correspond_keys_p->table_idx == roop_count
        ) {
          if (min_count > key_info2[tmp_correspond_keys_p->key_idx].key_parts)
          {
            min_count = key_info2[tmp_correspond_keys_p->key_idx].key_parts;
            correspond_pk[roop_count] = tmp_correspond_keys_p;
          }
          tmp_correspond_keys_p = tmp_correspond_keys_p->next;
        }
      }
    }

    for (roop_count = 0; roop_count < share->table_count; roop_count++)
      def_versions[roop_count] =
        part_tables[roop_count].table->s->get_table_def_version();

    if (share->zero_record_update_mode)
    {
      for (roop_count = 0; roop_count < share->table_count; roop_count++)
      {
        if (vp_bit_is_set(share->select_ignore_with_lock, roop_count))
        {
          for (roop_count2 = 0; roop_count2 < bitmap_size; roop_count2++)
          {
            cpy_clm_bitmap[roop_count2] |= correspond_columns_bit[
              (roop_count * bitmap_size) + roop_count2];
          }
        }
      }
    }

    /* auto_increment column check */
    if (table->found_next_number_field)
    {
      part_table = part_tables[share->auto_increment_table].table;
      if (
        !part_table->found_next_number_field ||
        part_table->found_next_number_field->field_index !=
          correspond_columns_p[table_share->fields *
          share->auto_increment_table +
          table->found_next_number_field->field_index]
      ) {
        error_num = ER_VP_CANT_CORRESPOND_AUTO_INC_NUM;
        my_printf_error(error_num, ER_VP_CANT_CORRESPOND_AUTO_INC_STR, MYF(0),
          part_table_share->table_name.str);
        goto error_auto_inc_correspond;
      }
    }

    share->bitmap_size = bitmap_size;
    share->correspond_columns_p = correspond_columns_p;
    share->correspond_columns_c_ptr = correspond_columns_c_ptr;
    share->correspond_columns_bit = correspond_columns_bit + bitmap_size;
    share->all_columns_bit = correspond_columns_bit;
    share->keys = keys;
    share->largest_key = largest_key;
    share->correspond_pk = correspond_pk;
    share->correspond_keys_p_ptr = correspond_keys_p_ptr;
    share->cpy_clm_bitmap = cpy_clm_bitmap;
    share->need_converting = need_converting;
    share->need_searching = need_searching;
    share->need_full_col_for_update = need_full_col_for_update;
    share->pk_in_read_index = pk_in_read_index;
    share->def_versions = def_versions;
    share->init = TRUE;
    share->reinit = FALSE;
  }
  pthread_mutex_unlock(&share->init_mutex);
  DBUG_RETURN(0);

error_auto_inc_correspond:
error_key_correspond:
error_column_correspond:
error_table_correspond:
  my_free(correspond_columns_p, MYF(0));
error_alloc:
  pthread_mutex_unlock(&share->init_mutex);
  DBUG_RETURN(error_num);
}

uchar vp_bit_count(
  uchar bitmap
) {
  DBUG_ENTER("vp_bit_count");
  DBUG_PRINT("info",("vp bitmap=%d", bitmap));
  bitmap = ((bitmap & 0xaa) >> 1) + (bitmap & 0x55);
  bitmap = ((bitmap & 0xcc) >> 2) + (bitmap & 0x33);
  bitmap = ((bitmap & 0xf0) >> 4) + (bitmap & 0x0f);
  DBUG_PRINT("info",("vp bitcount=%d", bitmap));
  DBUG_RETURN(bitmap);
}

void *vp_bg_action(
  void *arg
) {
  VP_BG_BASE *base = (VP_BG_BASE *) arg;
  ha_vp *vp = (ha_vp *) base->parent;
  TABLE *table;
  THD *thd;
  my_thread_init();
  DBUG_ENTER("vp_bg_action");
  /* init start */
  if (!(thd = new THD()))
  {
    pthread_mutex_lock(&base->bg_sync_mutex);
    pthread_cond_signal(&base->bg_sync_cond);
    pthread_mutex_unlock(&base->bg_sync_mutex);
    my_thread_end();
    DBUG_RETURN(NULL);
  }
  thd->thread_stack = (char*) &thd;
  thd->store_globals();
  base->bg_thd = thd;
  pthread_mutex_lock(&base->bg_mutex);
  pthread_mutex_lock(&base->bg_sync_mutex);
  pthread_cond_signal(&base->bg_sync_cond);
  base->bg_init = TRUE;
  pthread_mutex_unlock(&base->bg_sync_mutex);
  /* init end */

  while (TRUE)
  {
    pthread_cond_wait(&base->bg_cond, &base->bg_mutex);
    DBUG_PRINT("info",("vp bg loop start"));
    if (base->bg_caller_sync_wait)
    {
      pthread_mutex_lock(&base->bg_sync_mutex);
      pthread_cond_signal(&base->bg_sync_cond);
      pthread_mutex_unlock(&base->bg_sync_mutex);
    }
    switch (base->bg_command)
    {
      case VP_BG_COMMAND_KILL:
        DBUG_PRINT("info",("vp bg kill start"));
        pthread_mutex_lock(&base->bg_sync_mutex);
        pthread_cond_signal(&base->bg_sync_cond);
        pthread_mutex_unlock(&base->bg_mutex);
        pthread_mutex_unlock(&base->bg_sync_mutex);
        delete thd;
#if !defined(MYSQL_DYNAMIC_PLUGIN) || !defined(_WIN32)
        my_pthread_setspecific_ptr(THR_THD, NULL);
#endif
        my_thread_end();
        DBUG_RETURN(NULL);
      case VP_BG_COMMAND_SELECT:
        DBUG_PRINT("info",("vp bg select start"));
        table = base->part_table->table;
#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 50200
        base->bg_error = table->file->ha_index_read_map(
          table->record[0], (const uchar *) base->table_key,
          base->tgt_key_part_map, HA_READ_KEY_EXACT);
#else
        base->bg_error = table->file->index_read_map(
          table->record[0], (const uchar *) base->table_key,
          base->tgt_key_part_map, HA_READ_KEY_EXACT);
#endif
        break;
      case VP_BG_COMMAND_INSERT:
        DBUG_PRINT("info",("vp bg insert start"));
        table = base->part_table->table;
/*
        table->next_number_field = table->found_next_number_field;
*/
        base->bg_error = table->file->ha_write_row(table->record[0]);
        table->next_number_field = NULL;
        table->auto_increment_field_not_null = FALSE;
        break;
      case VP_BG_COMMAND_UPDATE:
        table = base->part_table->table;
        base->bg_error = table->file->ha_update_row(
          table->record[1], table->record[0]);
        break;
      case VP_BG_COMMAND_DELETE:
        table = base->part_table->table;
        base->bg_error = table->file->ha_delete_row(table->record[0]);
        break;
      case VP_BG_COMMAND_UPDATE_SELECT:
        DBUG_PRINT("info",("vp bg update select start"));
        table = base->part_table->table;
#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 50200
        base->bg_error = table->file->ha_index_read_idx_map(
          table->record[0], base->key_idx, (const uchar *) base->table_key,
          base->tgt_key_part_map, HA_READ_KEY_EXACT);
#else
        base->bg_error = table->file->index_read_idx_map(
          table->record[0], base->key_idx, (const uchar *) base->table_key,
          base->tgt_key_part_map, HA_READ_KEY_EXACT);
#endif
        if (!base->bg_error)
        {
          if (base->record_idx)
          {
            store_record(table, record[1]);
          } else {
            base->bg_error = table->file->ha_delete_row(table->record[0]);
          }
        }
        break;
    }
    continue;
  }
}

int vp_table_num_list_to_bitmap(
  VP_SHARE *share,
  char *table_num_list,
  uchar *bitmap
) {
  int table_idx;
  DBUG_ENTER("vp_table_num_list_to_bitmap");
  while(*table_num_list)
  {
    while (*table_num_list == ' ')
      table_num_list++;
    if (*table_num_list)
    {
      table_idx = atoi(table_num_list);
      if (table_idx > share->table_count || table_idx < 1)
      {
        my_printf_error(ER_VP_TBL_NUM_OUT_OF_RANGE_NUM,
          ER_VP_TBL_NUM_OUT_OF_RANGE_STR, MYF(0), table_idx);
        DBUG_RETURN(ER_VP_TBL_NUM_OUT_OF_RANGE_NUM);
      }
      vp_set_bit(bitmap, table_idx - 1);
    }
    if (!(table_num_list = strchr(table_num_list, ' ')))
      break;
  }
  DBUG_RETURN(0);
}
